import time
import logging

import typing as tp

from IPython.display import display
from copy import deepcopy
from typing import List, Optional, Any, Union

from .ipythonwidget import MetricWidget

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class MetricsWidget(MetricWidget):
    def __init__(self):
        super(self.__class__, self).__init__()

    def update_data(self, data: tp.Dict) -> None:
        # deepcopy is crucial here
        self.data = deepcopy(data)


class MetricsPlotter:
    """
    Context manager that enables widget with learning curves in
    JupyterLab / Jupyter Notebook
    """

    def __init__(self, train_metrics: List[Union[str, tp.Dict[str, str]]],
                 test_metrics: Optional[List[Union[str, tp.Dict[str, str]]]] = None,
                 total_iterations: Optional[int] = None) -> None:
        """
        Constructor that defines metrics to be plotted and total iterations count.

        Parameters
        ----------
        train_metrics : list of str or list of dict
            List of train metrics to be tracked.
            Each item in list can be either string with metric name or dict
            with the following format:
            {
                "name": "{metric_name}",
                "best_value": "Max|Min|Undefined",
            }

        test_metrics : list of str or list of dict, optional (default=None)
            List of test metrics to be tracked.
            Has the same format as train_metrics. Equals to train_metrics, if not defined

        total_iterations: int, optional (default=None)
            Total number of iterations, allows for remaining time estimation.
        """

        self._widget = MetricsWidget()

        self._values = {
            "iterations": [],
            "meta": {
                "launch_mode": "Train",
                "parameters": "",
                "name": "experiment",
                "iteration_count": None,  # set later
                "learn_sets": ["learn"],
                "learn_metrics": None,  # set later
                "test_sets": ["test"],
                "test_metrics": None,  # set later
            }
        }

        self._content = {
            "passed_iterations": 0,
            "total_iterations": None,  # set later
            "data": self._values,
        }

        # data propagated to widget class
        self._data = {
            "test_path": {
                "path": "test_path",
                "name": "experiment",
                "content": self._content,
            }
        }

        test_metrics = test_metrics or train_metrics
        train_metrics_meta: List[tp.Dict[str, str]] = self.construct_metrics_meta(train_metrics)
        test_metrics_meta: List[tp.Dict[str, str]] = self.construct_metrics_meta(test_metrics)

        self._train_metrics_positions = {
            meta["name"]: pos for pos, meta in enumerate(train_metrics_meta)}
        self._test_metrics_positions = {
            meta["name"]: pos for pos, meta in enumerate(test_metrics_meta)}

        self._values["meta"].update({
            "learn_metrics": train_metrics_meta,
            "test_metrics": test_metrics_meta,
        })

        self.passed_iterations = 0
        self.total_iterations = total_iterations

        if total_iterations is not None:
            self._values["meta"]["iteration_count"] = total_iterations

        self._content["total_iterations"] = total_iterations or 0

        self._start_time = time.time()

    def __enter__(self) -> 'MetricsPlotter':
        display(self._widget)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> Any:
        if exc_type == KeyboardInterrupt:
            logger.info(
                f"Learning was stopped manually after {self.passed_iterations} epochs")
            return True

    @staticmethod
    def construct_metrics_meta(metrics: List[Union[str, tp.Dict[str, str]]]) -> List[tp.Dict[str, str]]:
        meta: List[tp.Dict[str, str]] = []
        for item in metrics:
            if isinstance(item, str):
                name, best_value = item, "Undefined"
            elif isinstance(item, dict):
                assert "name" in item and "best_value" in item, \
                    "Wrong metrics definition format: should have " \
                    "`name` and `best_value` fields"
                name, best_value = item["name"], item["best_value"]
            else:
                assert False, "Each metric should be defined as str or as" \
                    "dict with `name` and `best_value` fields"
            meta.append({"best_value": best_value, "name": name})
        return meta

    @staticmethod
    def construct_metrics_array(metrics_positions: tp.Dict[str, int],
                                metrics: tp.Dict[str, float]) -> List[float]:
        array: List[float] = [0.] * len(metrics_positions)

        # data validation
        assert set(metrics.keys()) == set(metrics_positions.keys()), \
            f"Not all metrics were passed while logging, expected " \
            f"following: {', '.join(list(metrics_positions.keys()))}"

        for metric, value in metrics.items():
            assert isinstance(value, float), "Type of metric {metric} should be float"
            array[metrics_positions[metric]] = value
        return array

    def estimate_remaining_time(self, time_from_start: float) -> Optional[float]:
        if self.total_iterations is None:
            return None
        remaining_iterations: int = self.total_iterations - self.passed_iterations
        return time_from_start / self.passed_iterations * remaining_iterations

    def log(self, epoch: int, train: bool, metrics: tp.Dict[str, float]) -> None:
        """
        Save metrics at specific training epoch.

        Parameters
        ----------
        epoch : int
            Current epoch

        train : bool
            Flag that indicates whether metrics are calculated on train or test data

        metrics: dict
            Values for each of metrics defined in `__init__` method of this class
        """

        self.passed_iterations = epoch + 1
        self._content["passed_iterations"] = self.passed_iterations
        total_iterations = max(self._content["total_iterations"], self.passed_iterations)

        self._content["total_iterations"] = total_iterations
        self._values["meta"]["iteration_count"] = total_iterations

        assert len(self._values["iterations"]) in (epoch, epoch + 1), \
            "Data for epochs should be passed successively (wrong epoch number)"

        time_from_start: float = time.time() - self._start_time

        should_redraw: bool = len(self._values["iterations"]) == epoch + 1

        if len(self._values["iterations"]) == epoch:
            self._values["iterations"].append({
                "learn": [],
                "test": [],
                "iteration": epoch,
                "passed_time": time_from_start,
            })

            remaining_time = self.estimate_remaining_time(time_from_start)
            if remaining_time is not None:
                self._values["iterations"][-1]["remaining_time"] = remaining_time

        key: str = "learn" if train else "test"
        value: List[float] = self.construct_metrics_array(
            self._train_metrics_positions if train else self._test_metrics_positions, metrics)

        self._values["iterations"][-1].update({key: value})

        if should_redraw:
            self._widget.update_data(self._data)



import logging

filename = None
eval_logger = None
console_handler = None


def init():
    global eval_logger
    eval_logger_name = "eval_feature"
    eval_logger = logging.getLogger(eval_logger_name)
    eval_logger.setLevel(logging.CRITICAL)
    global console_handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.CRITICAL)

    formatter = logging.Formatter("[%(levelname)s]: %(message)s")
    console_handler.setFormatter(formatter)

    eval_logger.addHandler(console_handler)


def set_logger_name(name):
    global filename
    filename = name


def get_eval_logger():
    if eval_logger is None:
        init()
    return eval_logger


def set_level(level):
    if eval_logger is None:
        init()
    console_handler.setLevel(level)



import hashlib
import logging
import numpy as np
import os
import pandas as pd
import tarfile
import tempfile
import six
import shutil

from .core import PATH_TYPES, fspath


logger = logging.getLogger(__name__)


def _extract(src_file, dst_dir='.'):
    cur_dir = os.getcwd()
    os.chdir(dst_dir)
    try:
        with tarfile.open(src_file, 'r:gz') as f:
            f.extractall()
    finally:
        os.chdir(cur_dir)


def _calc_md5(path):
    hasher = hashlib.md5()
    with open(path, 'rb') as f:
        while True:
            block = f.read(65536)
            if not block:
                break
            hasher.update(block)
    return hasher.hexdigest()


def _ensure_dir_exists(path):
    try:
        os.makedirs(path)
    except OSError:
        if not os.path.isdir(path):
            raise


def _cached_download(url, md5, dst):
    if os.path.isfile(dst) and _calc_md5(dst) == md5:
        return

    def reporthook(blocknum, bs, size):
        logger.debug('downloaded %s bytes', size)

    urls = url if isinstance(url, list) or isinstance(url, tuple) else (url, )

    for u in urls:
        try:
            six.moves.urllib.request.urlretrieve(u, dst, reporthook=reporthook)
            break
        except (six.moves.urllib.error.URLError, IOError):
            logger.debug('failed to download from %s', u)
    else:
        raise RuntimeError('failed to download from %s', urls)

    dst_md5 = _calc_md5(dst)
    if dst_md5 != md5:
        raise RuntimeError('md5 sum mismatch for url {url}; expected {expected}, but got {got}'.format(
            url=u, expected=md5, got=dst_md5))


_cache_path = None


def _get_cache_path():
    global _cache_path
    if _cache_path is None:
        _cache_path = os.path.join(os.getcwd(), 'catboost_cached_datasets')
    return _cache_path


def set_cache_path(path):
    assert isinstance(path, PATH_TYPES), 'expected string or pathlib.Path'
    global _cache_path
    _cache_path = fspath(path)


def _download_dataset(url, md5, dataset_name, train_file, test_file, cache=False):
    # TODO(yazevnul): this is not thread safe (or process safe?), we should take a file lock when
    # enter this function to avoid dataset being overwritten or corrupted or something else that may
    # have happen when OS operated simultaneously on the same file. Same thing should probably be
    # done with `_cached_download`.
    dir_path = os.path.join(_get_cache_path(), dataset_name) if cache else tempfile.mkdtemp()
    train_path = os.path.join(dir_path, train_file)
    test_path = os.path.join(dir_path, test_file)
    if not (os.path.exists(train_path) and os.path.exists(test_path)):
        _ensure_dir_exists(dir_path)
        file_descriptor, file_path = tempfile.mkstemp()
        os.close(file_descriptor)
        try:
            _cached_download(url, md5, file_path)
            _extract(file_path, dir_path)
        finally:
            os.remove(file_path)
    # move files for safe delete of temp dir
    if not cache:
        new_train_path = tempfile.mktemp()
        new_test_path = tempfile.mktemp()
        os.rename(train_path, new_train_path)
        os.rename(test_path, new_test_path)
        shutil.rmtree(dir_path)
        train_path, test_path = new_train_path, new_test_path
    return train_path, test_path


def _load_dataset_pd(url, md5, dataset_name, train_file, test_file, sep=',', header='infer', cache=False):
    train_path, test_path = _download_dataset(url, md5, dataset_name, train_file, test_file, cache)
    train, test = pd.read_csv(train_path, header=header, sep=sep), pd.read_csv(test_path, header=header, sep=sep)
    if not cache:
        os.remove(train_path)
        os.remove(test_path)
    return train, test


def _load_numeric_only_dataset(path, row_count, column_count, sep='\t'):
    # - can't use `pandas.read_csv` because it may result in 5x overhead
    # - can't use `numpy.loadtxt` because it may result in 3x overhead
    # And both mentioned above solutions are very slow compared to the one implemented below.
    dataset = np.zeros((row_count, column_count, ), dtype=np.float32, order='F')
    with open(path, 'rb') as f:
        for line_idx, line in enumerate(f):
            # `str.split()` is too slow, use `numpy.fromstring()`
            row = np.fromstring(line, dtype=np.float32, sep=sep)
            assert row.size == column_count, 'got too many columns at line %d (expected %d columns, got %d)' % (line_idx + 1, column_count, row.size)
            # doing `dataset[line_idx][:]` instead of `dataset[line_idx]` is here on purpose,
            # otherwise we may reallocate memory, while here we just copy
            dataset[line_idx][:] = row

    assert line_idx + 1 == row_count, 'got too many lines (expected %d lines, got %d)' % (row_count, line_idx + 1)

    return pd.DataFrame(dataset)


def titanic():
    url = 'https://storage.mds.yandex.net/get-devtools-opensource/233854/titanic.tar.gz'
    md5 = '9c8bc61d545c6af244a1d37494df3fc3'
    dataset_name, train_file, test_file = 'titanic', 'train.csv', 'test.csv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file)


def amazon():
    url = 'https://storage.mds.yandex.net/get-devtools-opensource/250854/amazon.tar.gz'
    md5 = '8fe3eec12bfd9c4c532b24a181d0aa2c'
    dataset_name, train_file, test_file = 'amazon', 'train.csv', 'test.csv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file)


def msrank():
    url = 'https://storage.mds.yandex.net/get-devtools-opensource/233854/msrank.tar.gz'
    md5 = '34fee225d02419adc106581f4eb36f2e'
    dataset_name, train_file, test_file = 'msrank', 'train.tsv', 'test.tsv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file, header=None, sep='\t', cache=True)


def msrank_10k():
    url = 'https://storage.mds.yandex.net/get-devtools-opensource/250854/msrank_10k.tar.gz'
    md5 = '79c5b67397289c4c8b367c1f34629eae'
    dataset_name, train_file, test_file = 'msrank_10k', 'train.csv', 'test.csv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file, header=None)


def rotten_tomatoes():
    """
    Contains information from kaggle [1], which is made available here under the Open Database License (ODbL) [2].

    Download "rotten_tomatoes" [1] data set.

    Will return two pandas.DataFrame-s, first with train part (rotten_tomatoes.data) and second with test part
    (rotten_tomatoes.test) of the dataset.

    NOTE: This is a preprocessed version of the dataset.

    [1]: https://www.kaggle.com/rpnuser8182/rotten-tomatoes
    [2]: https://opendatacommons.org/licenses/odbl/1-0/index.html
    """
    url = 'https://catboost-opensource.s3.yandex.net/rotten_tomatoes.tar.gz'
    md5 = 'a07fed612805ac9e17ced0d82a96add4'
    dataset_name, train_file, test_file = 'rotten_tomatoes', 'learn.tsv', 'test.tsv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file, sep='\t')


def imdb():
    url = 'https://catboost-opensource.s3.yandex.net/imdb.tar.gz'
    md5 = '0fd62578d631ac3d71a71c3e6ced6f8b'
    dataset_name, train_file, test_file = 'imdb', 'learn.tsv', 'test.tsv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file, sep='\t')


def epsilon():
    """
    Download "epsilon" [1] data set.

    Will return two pandas.DataFrame-s, first with train part (epsilon_normalized) and second with
    test part (epsilon_normalized.t) of the dataset. Object class will be located in the first
    column of dataset.

    NOTE: This is a preprocessed version of the dataset. It was converted from libsvm format into
    tsv (CatBoost doesn't support libsvm format out of the box).

    [1]: https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary.html#epsilon
    """
    urls = (
        'https://proxy.sandbox.yandex-team.ru/785711439',
        'https://storage.mds.yandex.net/get-devtools-opensource/250854/epsilon.tar.gz', )
    md5 = '5bbfac403ac673da7d7ee84bd532e973'
    dataset_name, train_file, test_file = 'epsilon', 'train.tsv', 'test.tsv'
    train_path, test_path = _download_dataset(urls, md5, dataset_name, train_file, test_file, cache=True)
    return (
        _load_numeric_only_dataset(train_path, 400000, 2001, sep='\t'),
        _load_numeric_only_dataset(test_path, 100000, 2001, sep='\t'))


def monotonic1():
    """
    Dataset with monotonic constraints.
    Can be used for poisson regression.
    Has several numerical and several categorical features.
    The first column contains target values. Columns with names Cat* contain categorical features.
    Columns with names Num* contain numerical features.

    Dataset also contains several numerical features, for which monotonic constraints must hold.
    For features in columns named MonotonicNeg*, if feature value decreases, then prediction value must not decrease.
    Thus, if there are two samples x1, x2 with all features being equal except
    for a monotonic negative feature M, such that x1[M] > x2[M], then the following inequality must
    hold for predictions: f(x1) <= f(x2)
    """
    url = 'https://storage.mds.yandex.net/get-devtools-opensource/479623/monotonic1.tar.gz'
    md5 = '1b9d8e15bc3fd6f1498e652e7fc4f4ca'
    dataset_name, train_file, test_file = 'monotonic1', 'train.tsv', 'test.tsv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file, sep='\t', cache=True)


def monotonic2():
    """
    Dataset with monotonic constraints.
    Can be used for regression.
    The first column contains target values.
    Other columns contain contain numerical features, for which monotonic constraints must hold.

    For features in columns named MonotonicNeg*, if feature value decreases, then prediction
    value must not decrease. Thus, if there are two samples x1, x2 with all features being
    equal except for a monotonic negative feature MNeg, such that x1[MNeg] > x2[MNeg], then
    the following inequality must hold for predictions: f(x1) <= f(x2)
    For features in columns named MonotonicPos*, if feature value decreases, then prediction
    value must not increase. Thus, if there are two samples x1, x2 with all features being
    equal except for a monotonic positive feature MPos, such that x1[MPos] > x2[MPos],
    then the following inequality must hold for predictions: f(x1) >= f(x2)
    """
    url = 'https://storage.mds.yandex.net/get-devtools-opensource/250854/monotonic2.tar.gz'
    md5 = 'ce559e212cb72c156269f6f9a641baca'
    dataset_name, train_file, test_file = 'monotonic2', 'train.tsv', 'test.tsv'
    return _load_dataset_pd(url, md5, dataset_name, train_file, test_file, sep='\t')


def adult():
    """
    Download "Adult Data Set" [1] from UCI Machine Learning Repository.

    Will return two pandas.DataFrame-s, first with train part (adult.data) and second with test part
    (adult.test) of the dataset.

    [1]: https://archive.ics.uci.edu/ml/datasets/Adult
    """
    # via https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.names
    names = (
        'age', 'workclass', 'fnlwgt', 'education', 'education-num', 'marital-status', 'occupation',
        'relationship', 'race', 'sex', 'capital-gain', 'capital-loss', 'hours-per-week',
        'native-country', 'income', )
    dtype = {
        'age': float, 'workclass': object, 'fnlwgt': float, 'education': object,
        'education-num': float, 'marital-status': object, 'occupation': object,
        'relationship': object, 'race': object, 'sex': object, 'capital-gain': float,
        'capital-loss': float, 'hours-per-week': float,
        'native-country': object, 'income': object, }

    # proxy.sandbox.yandex-team.ru is Yandex internal storage, we first try to download it from
    # internal storage to avoid putting too much pressure on UCI storage from our internal CI

    train_urls = (
        'https://proxy.sandbox.yandex-team.ru/779118052',
        'https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data', )
    train_md5 = '5d7c39d7b8804f071cdd1f2a7c460872'
    train_path = tempfile.mktemp()
    _cached_download(train_urls, train_md5, train_path)

    test_urls = (
        'https://proxy.sandbox.yandex-team.ru/779120000',
        'https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.test', )
    test_md5 = '35238206dfdf7f1fe215bbb874adecdc'
    test_path = tempfile.mktemp()
    _cached_download(test_urls, test_md5, test_path)

    train_df = pd.read_csv(train_path, names=names, header=None, sep=r',\s*', na_values=['?'], engine='python')
    os.remove(train_path)

    # lines in test part end with dot, thus we need to fix last column of the dataset
    test_df = pd.read_csv(test_path, names=names, header=None, sep=r',\s*', na_values=['?'], skiprows=1, converters={'income': lambda x: x[:-1]}, engine='python')
    os.remove(test_path)

    # pandas 0.19.1 doesn't support `dtype` parameter for `read_csv` when `python` engine is used,
    # so we have to do the casting manually; also we can't use `converters` together with `dtype`
    train_df = train_df.astype(dtype)
    test_df = test_df.astype(dtype)

    return train_df, test_df


def higgs():
    """
    Download "higgs" [1] data set.

    Will return two pandas.DataFrame-s, first with train part and second with
    test part of the dataset. Object class will be located in the first
    column of dataset.

    [1]: https://archive.ics.uci.edu/ml/datasets/HIGGS
    """
    url = 'https://storage.mds.yandex.net/get-devtools-opensource/250854/higgs.tar.gz'
    md5 = 'ad59ba8328a9afa3837d7bf1a0e10e7b'
    dataset_name, train_file, test_file = 'higgs', 'train.tsv', 'test.tsv'
    train_path, test_path = _download_dataset(url, md5, dataset_name, train_file, test_file, cache=True)
    return (
        _load_numeric_only_dataset(train_path, 10500000, 29, sep='\t'),
        _load_numeric_only_dataset(test_path, 500000, 29, sep='\t'))



from . import _catboost

_carry_by_index = _catboost._carry_by_index
_carry_by_name = _catboost._carry_by_name
_uplift_by_index = _catboost._uplift_by_index
_uplift_by_name = _catboost._uplift_by_index


def carry(model, features):
    """
    Parameters
    ----------
    model :
        CatBoost / CatBoostClassifier / CatBoostRanker / CatBoostRegressor model
    features :
        must be a dict mapping strings (factor names) or integers (flat indexes) into floats
        NOTE: values in a dict can be lists of floats, but in this case they must all be the same size
    """
    model = model.copy()
    assert type(features) is dict

    factor_ids = []
    factor_values = []
    for factor_id, factor_value in features.items():
        assert type(factor_id) is str or type(factor_id) is int
        assert type(factor_value) is float or type(factor_value) is int or type(factor_value) is list

        if type(factor_value) is list:
            for value in factor_value:
                assert type(value) is float or type(value) is int
        else:
            factor_value = [factor_value]

        if len(factor_values):
            assert len(factor_values[0]) == len(factor_value)
            assert type(factor_ids[0]) == type(factor_id)

        factor_ids.append(factor_id)
        factor_values.append(factor_value)

    if len(factor_ids):
        if type(factor_ids[0]) is int:
            _carry_by_index(model._object, factor_ids, factor_values)
        else:
            _carry_by_name(model._object, factor_ids, factor_values)
    return model


def uplift(model, features):
    """
    Parameters
    ----------
    model :
        CatBoost / CatBoostClassifier / CatBoostRanker / CatBoostRegressor model
        NOTE: uplift allways use RawFormulaVal
    features :
        must be a dict mapping strings (factor names) or integers (flat indexes) into pairs of floats (base and next values)
    """
    model = model.copy()

    factor_ids = []
    factor_base_values = []
    factor_next_values = []
    for factor_id, (factor_base_value, factor_next_value) in features.items():
        assert type(factor_id) is str or type(factor_id) is int
        assert type(factor_base_value) is float or type(factor_base_value) is int
        assert type(factor_next_value) is float or type(factor_next_value) is int
        if len(factor_ids):
            assert type(factor_ids[0]) == type(factor_id)

        factor_ids.append(factor_id)
        factor_base_values.append(float(factor_base_value))
        factor_next_values.append(float(factor_next_value))

    if len(factor_ids):
        if type(factor_ids[0]) is str:
            _uplift_by_name(model._object, factor_ids, factor_base_values, factor_next_values)
        else:
            _uplift_by_index(model._object, factor_ids, factor_base_values, factor_next_values)

    return model



import json
import os
import warnings

from . import _catboost
fspath = _catboost.fspath


def try_plot_offline(figs):
    try:
        from plotly.offline import iplot
        from plotly.offline import init_notebook_mode
        init_notebook_mode(connected=True)
    except ImportError as e:
        warn_msg = "To draw plots you should install plotly."
        warnings.warn(warn_msg)
        raise ImportError(str(e))
    if not isinstance(figs, list):
        figs = [figs]
    for fig in figs:
        iplot(fig)


def save_plot_file(plot_file, plot_name, figs):
    warn_msg = "To draw plots you should install plotly."
    try:
        from plotly.offline import plot as plotly_plot
    except ImportError as e:
        warnings.warn(warn_msg)
        raise ImportError(str(e))

    def write_plot_file(plot_file_stream):
        plot_file_stream.write('\n'.join((
            '<html>',
            '<head>',
            '<meta charset="utf-8" />',
            '<script src="https://cdn.plot.ly/plotly-latest.min.js"></script>',
            '<title>{}</title>'.format(plot_name),
            '</head>',
            '<body>'
        )))
        for fig in figs:
            graph_div = plotly_plot(
                fig,
                output_type='div',
                show_link=False,
                include_plotlyjs=False
            )
            plot_file_stream.write('\n{}\n'.format(graph_div))
        plot_file_stream.write('</body>\n</html>')

    if isinstance(plot_file, str):
        with open(fspath(plot_file), 'w') as plot_file_stream:
            write_plot_file(plot_file_stream)
    else:
        write_plot_file(plot_file)


class OfflineMetricVisualizer(object):
    def __init__(self, train_dirs):
        if isinstance(train_dirs, str):
            train_dirs = [train_dirs]
        self._load_data(train_dirs)

    def _update_data_from_dir(self, path):
        data = {
            'iterations': [],
            'meta': {}
        }

        training_json = os.path.join(path, 'catboost_training.json')

        if os.path.isfile(training_json):
            try:
                with open(training_json, 'r') as json_data:
                    training_data = json.load(json_data)
                    data['meta'] = training_data['meta']
                    data['iterations'] = training_data['iterations']
            except ValueError:
                pass

        return {
            'passed_iterations': data['iterations'][-1]['iteration'] if data['iterations'] else 0,
            'total_iterations': data['meta']['iteration_count'] if data['meta'] else 0,
            'data': data
        }

    def _load_data(self, train_dirs):
        names = []
        curdir = os.path.abspath(os.path.curdir)
        for train_dir in train_dirs:
            abspath = os.path.abspath(train_dir)
            names.append(os.path.basename(abspath) if abspath != curdir else 'current')

        data = {}
        dirs = [{'name': name, 'path': path} for name, path in zip(names, train_dirs)]

        for dir_info in dirs:
            path = dir_info.get('path')
            content = self._update_data_from_dir(path)

            if not content:
                continue

            data[path] = {
                'path': path,
                'name': dir_info.get('name'),
                'content': content
            }
        self.data = data

    def _get_plotly_figs(self, title):
        try:
            import plotly.graph_objs as go
        except ImportError as err:
            warnings.warn("To save plots to files you should install plotly.")
            raise ImportError(str(err))

        figs = []
        for path, dir_data in self.data.items():
            meta = dir_data['content']['data']['meta']

            has_parameters = meta['parameters'] == 'parameters'

            # name -> {'learn_sets': [(name, metric_idx)], 'test_sets': [(name, metric_idx)]}
            metrics = {}
            for i, learn_metric in enumerate(meta['learn_metrics']):
                metric_name = learn_metric['name']
                metrics.setdefault(metric_name, {})
                metrics[metric_name]['learn_sets'] = [(set_name, i) for set_name in meta['learn_sets']]
            for i, test_metric in enumerate(meta['test_metrics']):
                metric_name = test_metric['name']
                metrics.setdefault(metric_name, {})
                metrics[metric_name]['test_sets'] = [(set_name, i) for set_name in meta['test_sets']]

            iterations = dir_data['content']['data']['iterations']

            for metric_name, subsets in metrics.items():
                fig = go.Figure()

                figure_title = (title if (dir_data['name'] == 'catboost_info') else dir_data['name']) + ' : ' + metric_name
                fig['layout']['title'] = go.layout.Title(text=figure_title)

                learn_graph_color = 'rgb(160,0,0)'

                for learn_set_name, metric_idx in subsets.get('learn_sets', []):
                    fig.add_trace(go.Scatter(
                        x=[e['iteration'] for e in iterations],
                        y=[e[learn_set_name][metric_idx] for e in iterations],
                        line=go.scatter.Line(color=learn_graph_color),
                        mode='lines',
                        name=learn_set_name
                    ))

                test_graph_color = 'rgb(0,160,0)'

                for test_set_name, metric_idx in subsets.get('test_sets', []):

                    def generate_params_hover(e):
                        result = []
                        for param_name, param_value in e["parameters"][0].items():
                            result.append(param_name + ' : ' + str(param_value))
                        return '<br>'.join(result)

                    fig.add_trace(go.Scatter(
                        x=[e['iteration'] for e in iterations],
                        y=[e[test_set_name][metric_idx] for e in iterations],
                        line=go.scatter.Line(color=test_graph_color),
                        mode='lines',
                        name=test_set_name,
                        hovertext=[generate_params_hover(e) for e in iterations] if has_parameters else None
                    ))

                fig.update_layout(
                    xaxis=dict(title='iterations'),
                    yaxis=dict(title=metric_name)
                )

                figs.append(fig)

        return figs

    def save_to_file(self, title, file_output):
        save_plot_file(file_output, title, self._get_plotly_figs(title))



from . import _catboost

is_groupwise_metric = _catboost.is_groupwise_metric
is_multiclass_metric = _catboost.is_multiclass_metric
is_multitarget_objective = _catboost.is_multitarget_objective
is_classification_objective = _catboost.is_classification_objective
get_multi_quantile_approx_size = _catboost.get_multi_quantile_approx_size
is_pairwise_metric = _catboost.is_pairwise_metric
is_maximizable_metric = _catboost.is_maximizable_metric
is_minimizable_metric = _catboost.is_minimizable_metric
is_user_defined_metric = _catboost.is_user_defined_metric
has_gpu_implementation_metric = _catboost.has_gpu_implementation_metric
get_experiment_name = _catboost.get_experiment_name
convert_features_to_indices = _catboost.convert_features_to_indices
get_num_feature_values_sample = _catboost.get_num_feature_values_sample



try:
    from xgboost.callback import TrainingCallback as XGBTrainingCallback
except Exception:
    class XGBTrainingCallback:
        pass


from IPython.display import display

from .metrics_plotter import MetricsPlotter


class XGBPlottingCallback(XGBTrainingCallback):
    '''XGBoost callback with metrics plotting widget from CatBoost
    '''

    def __init__(self, total_iterations: int):
        self.plotter = None
        self.total_iterations = total_iterations

    def after_iteration(self, model, epoch, evals_log):

        data_names = evals_log.keys()
        # if more than one sample is passed, consider first as train sample
        first_train = (
            all(['valid' in data_name.lower() for data_name in data_names])
            and len(data_names) > 1
        )

        for data_name, metrics_info in evals_log.items():
            if "train" in data_name.lower() or first_train:
                train = True
                first_train = False
            elif "valid" in data_name.lower() or "test" in data_name.lower():
                train = False
            else:
                raise Exception("Unexpected sample name during evaluation")

            metrics = {name: values[-1] for name, values in metrics_info.items()}

            if self.plotter is None:
                names = list(metrics.keys())
                self.plotter = MetricsPlotter(names, names, self.total_iterations)
                display(self.plotter._widget)

            self.plotter.log(epoch, train, metrics)

        # False to indicate training should not stop.
        return False


def lgbm_plotting_callback():
    """LightGBM callback with metrics plotting widget from CatBoost
    """

    plotter = None

    def _init(env):
        train_metrics = []
        test_metrics = []

        for item in env.evaluation_result_list:
            assert len(item) == 4, "Plotting was run in not suppored mode"
            data_name, eval_name = item[:2]

            if "train" in data_name.lower():
                train_metrics.append(eval_name)
            elif "valid" in data_name.lower() or "test" in data_name.lower():
                test_metrics.append(eval_name)
            else:
                raise Exception("Unexpected sample name during evaluation")

        nonlocal plotter
        plotter = MetricsPlotter(
            train_metrics, test_metrics, env.end_iteration - env.begin_iteration)

        display(plotter._widget)

    def _callback(env):
        if plotter is None:
            _init(env)

        metrics = {"train": {}, "test": {}}

        for item in env.evaluation_result_list:
            data_name, eval_name, result = item[:3]

            if "train" in data_name.lower():
                metrics["train"][eval_name] = result
            elif "valid" in data_name.lower() or "test" in data_name.lower():
                metrics["test"][eval_name] = result
            else:
                raise Exception("Unexpected sample name during evaluation")

        plotter.log(
            env.iteration - env.begin_iteration,
            train=True,
            metrics=metrics["train"]
        )
        plotter.log(
            env.iteration - env.begin_iteration,
            train=False,
            metrics=metrics["test"]
        )

    _callback.order = 20
    return _callback



"""
This module contains the abstractions for keeping small amount of data. It provides the interface for classes and a few
realisations.
"""

from __future__ import print_function

import os

from .. import CatBoostError
from ..eval.log_config import get_eval_logger
from .utils import make_dirs_if_not_exists


# Interface
class FoldStorage(object):
    """
    Base class.
    """
    default_dir = 'folds'

    @staticmethod
    def remove_dir():
        """
        Remove default directory for folds if there're no files nut models. In other way it raises warning.

        Args:
            :return: Nothing.

        """
        try:
            if os.path.exists(_FoldFile.default_dir):
                os.rmdir(_FoldFile.default_dir)
        except OSError as err:
            get_eval_logger().warning(err.message)

    def __init__(self, fold, storage_name, sep, column_description):
        self._fold = fold
        self._storage_name = storage_name
        self._column_description = column_description
        self._sep = sep
        self._size = 0

    def get_separator(self):
        """
        Args:
            :return: (str) Delimiter for data used when we saved fold to file.

        """
        return self._sep

    def __str__(self):
        return self._storage_name

    def column_description(self):
        """
        Args:
            :return: (str) Path to the column description.

        """
        return self._column_description

    def contains_group_id(self, group_id):
        """
        Args:
            :param group_id: (int) The number of group we want to check.
            :return: True if fold contains line or lines with that group id.

        """
        return group_id in self._fold

    def open(self):
        raise NotImplementedError("The base class don't have delete method. Please, use successor.")

    def close(self):
        raise NotImplementedError("The base class don't have delete method. Please, use successor.")

    def delete(self):
        raise NotImplementedError("The base class don't have delete method. Please, use successor.")


class _FoldFile(FoldStorage):
    """
    FoldFile is the realisation of the interface of FoldStorage. It always saves data to file before reset them.
    All files place to the special directory 'folds'.
    """

    def __init__(self, fold, storage_name, sep, column_description):
        super(_FoldFile, self).__init__(
            fold, storage_name,
            sep=sep, column_description=column_description
        )
        self._file_path = os.path.join(self.default_dir,
                                       storage_name)
        self._prepare_path()
        self._lines = []
        self._file = None

    def _prepare_path(self):
        make_dirs_if_not_exists(self.default_dir)
        open(self._file_path, 'w').close()  # clean file

    def path(self):
        return self._file_path

    def add(self, line):
        self._size += 1
        print(line, file=self._file, end='')

    def add_all(self, lines):
        [self.add(line) for line in lines]

    def open(self):
        if self._file is None:
            self._file = open(self._file_path, mode='a')
        else:
            raise CatBoostError("File already opened {}".format(self._file_path))

    def is_opened(self):
        return self._file is not None

    def close(self):
        if self._file is None:
            raise CatBoostError("Trying to close None {}".format(self._file_path))

        self._file.close()
        self._file = None

    def delete(self):
        if self._file is not None:
            raise CatBoostError("Close file before delete")

        if os.path.exists(self._file_path):
            os.remove(self._file_path)
